"""
Special string types
"""
from enum import Enum
from typing import TypeVar

from flags import Flags
from mhelper.exception_helper import SwitchError
from mhelper.generics_helper import MAnnotationFactory, MAnnotation


T = TypeVar( "T" )


class NoDefaultType:
    """
    Type of `NOT_PROVIDED` constant. Used to distinguish a value of `None` from a value that isn't provided.
    """
    def __str__(self) -> str:
        return "(mandatory)"


NOT_PROVIDED = NoDefaultType( )
"""
NOT_PROVIDED is used to distinguish between a value of `None` and a value that that isn't even provided.

As a parameter default, it can be considered to mean "mandatory".
When passed as a parameter's value, it can be considered to mean "I'm not specifying this, use your default (or raise an Exception if there is no default)."
"""


# ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒
# ▒ ENUMS ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒
# ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒



class MEnum( Enum ):
    """
    An enum class that presents a less useless string function
    """
    
    
    def __str__( self ):
        return self.name


class MFlags( Flags ):
    """
    A flags class that presents  less useless string function
    """
    __no_flags_name__ = "NONE"
    __all_flags_name__ = "ALL"
    
    
    def __str__( self ):
        return self.to_simple_str()


# ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒
# ▒ FILE NAME ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒
# ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒



class EFileMode( MEnum ):
    UNKNOWN = 0
    READ = 1
    WRITE = 2


_default_filename_path = None


class FileNameAnnotation( MAnnotation ):
    def __init__( self, args ):
        super().__init__( args )
        
        mode = EFileMode.UNKNOWN
        ext = None
        
        for p in self.parameters:
            if isinstance( p, EFileMode ):
                mode = p
            elif p == "r":
                mode = EFileMode.READ
            elif p == "w":
                mode = EFileMode.WRITE
            elif isinstance( p, str ):
                ext = p
            else:
                raise SwitchError( "parameter", p, instance = True )
        
        self.mode = mode
        self.extension = ext
        
    def __str__(self):
        assert self.child is str
        return "{}[{} {}]".format( self.factory, self.mode.name, self.extension )


Filename : FileNameAnnotation = MAnnotationFactory( "Filename", annotation_type = FileNameAnnotation )[str]
"""File names"""

Dirname = MAnnotationFactory( "Dirname" )[str]
"""Directory names"""

MOptional = MAnnotationFactory( "MOptional" )
"""MOptional is Optional, but doesn't complain about the argument not being a type.
It will be redundant when typing"""

MUnion = MAnnotationFactory( "MOptional" )
"""MOptional is Optional, but doesn't complain about the argument not being a type.
It will be redundant when typing"""


# ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒
# ▒ PASSWORD  ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒
# ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒



class Password:
    """
    Passwords
    """
    
    
    def __init__( self, value: str ):
        self.__value = value
    
    
    @property
    def value( self ):
        return self.__value
    
    
    def __str__( self ) -> str:
        return "********"


# ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒
# ▒ READ-ONLY ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒
# ▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒


HReadonly = MAnnotationFactory( "Readonly" )



"""
Functions for dealing with ANSI strings.
"""
from typing import Iterator, cast, Optional, Iterable
import re

from collections import defaultdict

from mhelper import string_helper, SwitchError, exception_helper


_strip_ansi_rx = re.compile( r'(\x1b[^m]+m)' )
_AnsiStr_ = "AnsiStr"


class AnsiStr:
    """
    Manages an ANSI string as if it were a normal string.
    
    The codes are bound to prefix the characters before which they appear, hence where "#" are ANSI codes such as RED, BLUE, YELLOW, RESET...
    
    ```
    a = "#hello#.#world#"
    a[:8] == "#hello#.#wo"
    a[0] == "#h"
    a[1] == "e"
    a[-1] == "d#"
    ```
    
    The exception is any trailing code, which is bound after the last character.
    """
    
    
    def __init__( self, with_ansi: str ):
        exception_helper.assert_type( "with_ansi", with_ansi, str )
        
        self.with_ansi = with_ansi
        self.ansi_positions = defaultdict( str )
        
        elements = self.get_elements()
        
        texts = [elements[0]]
        c = len( elements[0] )
        
        for i in range( 1, len( elements ), 2 ):
            ansi = elements[i]
            
            self.ansi_positions[c] += ansi
            
            text = elements[i + 1]
            c += len( text )
            
            texts.append( text )
        
        self.without_ansi = "".join( texts )
        self.length = len( self.without_ansi )
    
    
    def __add__( self, other: str ) -> _AnsiStr_:
        if isinstance( other, str ):
            return AnsiStr( self.with_ansi + other )
        elif isinstance( other, AnsiStr ):
            return AnsiStr( self.with_ansi + other.with_ansi )
        else:
            raise SwitchError( "other", other, instance = True )
    
    
    def __mul__( self, other ):
        return AnsiStr( self.with_ansi * other )
    
    
    def get_elements( self ):
        return _strip_ansi_rx.split( self.with_ansi )
    
    
    def lstrip( self, chars: Optional[str] = None ) -> _AnsiStr_:
        remove = len( self.without_ansi ) - len( self.without_ansi.lstrip( chars ) )
        return self[remove:]
    
    
    def rstrip( self, chars: Optional[str] = None ) -> _AnsiStr_:
        remove = len( self.without_ansi ) - len( self.without_ansi.rstrip( chars ) )
        return self[:-remove]
    
    
    def strip( self, chars: Optional[str] = None ) -> _AnsiStr_:
        remove_left = len( self.without_ansi ) - len( self.without_ansi.lstrip( chars ) )
        remove_right = len( self.without_ansi ) - len( self.without_ansi.rstrip( chars ) )
        
        if remove_left + remove_right >= len( self ):
            return AnsiStr( "" )
        
        return self[remove_left:-remove_right]
    
    
    def join( self, others: Iterable[str] ):
        text = self.with_ansi.join( str( x ) for x in others )
        return AnsiStr( text )
    
    
    def __str__( self ):
        return self.with_ansi
    
    
    def __repr__( self ):
        return "".join( self )
    
    
    def __len__( self ):
        return self.length
    
    
    def __iter__( self ):
        for i in range( len( self ) ):
            ansi = self.ansi_positions.get( i ) or ""
            
            yield "{}_{}".format( repr( self.without_ansi[i] ), repr( ansi ) )
        
        ansi = self.ansi_positions.get( len( self ) ) or ""
        yield "_+{}".format( repr( ansi ) )
    
    
    def __getitem__( self, key ):
        if isinstance( key, slice ):
            r = []
            
            for i in range( key.start or 0, key.stop or len( self ), key.step or 1 ):
                r.append( self[i] )
            
            return AnsiStr( "" ).join( r )
        
        if key < 0:
            key += len( self )
        
        if key == len( self ) - 1:
            suffix = self.ansi_positions.get( key + 1 )
        else:
            suffix = ""
        
        prefix = self.ansi_positions.get( key ) or ""
        text = self.without_ansi[key]
        return AnsiStr( prefix + text + suffix )
    
    
    def wrap( self, width: int ) -> Iterator[_AnsiStr_]:
        return cast( Iterator[AnsiStr], string_helper.wrap( self._str, width ) )
    
    
    def ljust( self, width: int, char: str = None ) -> _AnsiStr_:
        return cast( AnsiStr, string_helper.ljust( self._str, width, self.__spacer( char )._str ) )
    
    
    def rjust( self, width: int, char: str = None ) -> _AnsiStr_:
        return cast( AnsiStr, string_helper.rjust( self._str, width, self.__spacer( char )._str ) )
    
    
    def cjust( self, width: int, char: str = None ) -> _AnsiStr_:
        return cast( AnsiStr, string_helper.cjust( self._str, width, self.__spacer( char )._str ) )
    
    
    def fix_width( self, width: int, char: str = None, justify : int = None ) -> _AnsiStr_:
        return cast( AnsiStr, string_helper.fix_width( self._str, width, self.__spacer( char )._str, justify ) )
    
    
    @property
    def _str( self ) -> str:
        return cast( str, self )
    
    
    @classmethod
    def __spacer( cls, char: str ) -> _AnsiStr_:
        if char is None:
            return _SPACE
        elif isinstance( char, AnsiStr ):
            return char
        else:
            return AnsiStr( char )


_SPACE = AnsiStr( " " )


def without_ansi( text: str ) -> str:
    return AnsiStr( text ).without_ansi


def length( text: str ) -> int:
    return len( AnsiStr( text ) )


def wrap( text: str, width: int ) -> Iterator[str]:
    return (str( x ) for x in AnsiStr( text ).wrap( width ))


def ljust( text: str, width: int, char = None ) -> str:
    return str( AnsiStr( text ).ljust( width, char ) )


def rjust( text: str, width: int, char = None ) -> str:
    return str( AnsiStr( text ).rjust( width, char ) )


def cjust( text: str, width: int, char: str = None ) -> str:
    return str( AnsiStr( text ).cjust( width, char ) )


def fix_width( text: str, width: int, char: str = None, justify : int = None ) -> str:
    return str( AnsiStr( text ).fix_width( width, char, justify ) )



"""
Functions for formatting stuff using ANSI codes and/or esoteric UNICODE characters.
"""

from typing import Union, Iterable

from colorama import Back, Fore, Style

from mhelper import ansi, ansi_helper, exception_helper, string_helper


def format_source( text: str,
                   keywords: Iterable[str],
                   variables: Iterable[str] ) -> str:
    """
    Prints source text, highlighting keywords and variables, and prefixing line numbers
    
    :param text:        Text to print
    :param keywords:    Keywords to highlight
    :param variables:   Variables to highlight
    :return:            Nothing
    """
    r = []
    
    text = text.split( "\n" )
    
    for i, line in enumerate( text ):
        prefix = Back.LIGHTBLACK_EX + Fore.BLACK + " " + str( i ).rjust( 4 ) + " " + Style.RESET_ALL + " "
        
        line = string_helper.highlight_words( line, keywords, Style.RESET_ALL + Fore.GREEN, Style.RESET_ALL )
        line = string_helper.highlight_words( line, variables, Style.RESET_ALL + Fore.CYAN, Style.RESET_ALL )
        
        r.append( prefix + line )
    
    return "\n".join( r )


def format_traceback( exception: Union[BaseException, str],
                      traceback_ = None,
                      warning = False,
                      wordwrap = 0 ) -> str:
    """
    Formats a traceback.
    
    :param exception:       Exception to display 
    :param traceback_:      Traceback text (leave as `None` to get the system traceback) 
    :param warning:         Set to `True` to use warning, rather than error, colours 
    :param wordwrap:        Set to the wordwrap width. 
    :return:                ANSI containing string  
    """
    r = []
    
    wordwrap = wordwrap or 120
    
    if warning:
        MAIN_LINE = Style.RESET_ALL + Back.YELLOW + Fore.BLACK
        MAIN_LINE_H = Style.RESET_ALL + Style.BRIGHT + Back.YELLOW + Fore.BLACK
        PREFIX = Back.LIGHTYELLOW_EX + Fore.BLACK
        NORMAL = Fore.CYAN
        CODE = Fore.CYAN + Style.DIM
        FILE_LINES = Style.DIM + Fore.CYAN
        FILE_NAMES = Fore.LIGHTCYAN_EX
    else:
        MAIN_LINE = Style.RESET_ALL + Back.RED + Fore.LIGHTWHITE_EX
        MAIN_LINE_H = ansi.RESET + ansi.BOLD + ansi.BACK_RED + ansi.FORE_BRIGHT_WHITE + ansi.ITALIC
        PREFIX = Back.LIGHTRED_EX + Fore.BLACK
        NORMAL = Fore.CYAN
        CODE = Style.RESET_ALL + Fore.CYAN + Style.DIM
        FILE_LINES = Style.RESET_ALL + Style.DIM + Fore.BLUE
        FILE_NAMES = Style.BRIGHT + Fore.YELLOW
    
    r.append( PREFIX + "┌" + "─" * (wordwrap - 2) + "┐" + Style.RESET_ALL )
    
    if isinstance( exception, BaseException ):
        ex = exception
        
        while ex:
            if ex is not exception:
                r.append( PREFIX + "│" + MAIN_LINE + " " + "---CAUSED BY---" + Style.RESET_ALL )
            
            r.append( PREFIX + "│" + MAIN_LINE + " *" + type( ex ).__name__ + "*" + Style.RESET_ALL )
            ex_text = MAIN_LINE + string_helper.highlight_quotes( str( ex ), "«", "»", MAIN_LINE_H, MAIN_LINE )
            
            for line in ansi_helper.wrap( ex_text, wordwrap ):
                r.append( PREFIX + "│" + MAIN_LINE + " " + line + Style.RESET_ALL )
            
            ex = ex.__cause__
    
    else:
        r.append( str( exception ) )
    
    if not traceback_:
        traceback_ = exception_helper.full_traceback()
    
    lines = traceback_.split( "\n" )
    
    for i, line in enumerate( lines ):
        print_line = line.strip()
        prefix = PREFIX + "│" + Style.RESET_ALL + " "
        
        if print_line.startswith( "File " ):
            print_line = FILE_LINES + string_helper.highlight_regex( print_line, "\\/([^\\/]*)\"", FILE_NAMES, FILE_LINES )
            print_line = FILE_LINES + string_helper.highlight_regex( print_line, "line ([0-9]*),", FILE_NAMES, FILE_LINES )
            print_line = FILE_LINES + string_helper.highlight_regex( print_line, "in (.*)$", CODE, FILE_LINES )
            
            r.append( prefix + print_line + Style.RESET_ALL )
        elif line.startswith( " " ):
            r.append( prefix + "    " + CODE + ansi.ITALIC + print_line + Style.RESET_ALL )
        elif line.startswith( "*" ):
            c = wordwrap - len( print_line ) - 6
            r.append( PREFIX + "├────" + print_line + "─" * c + "┤" + Style.RESET_ALL )
        else:
            r.append( prefix + NORMAL + print_line + Style.RESET_ALL )
    
    r.append( PREFIX + "└" + "─" * (wordwrap - 2) + "┘" + Style.RESET_ALL )
    
    return "\n".join( r )


def format_two_columns( left_margin: int,
                        centre_margin: int,
                        right_margin: int,
                        left_text: str,
                        right_text: str ):
    """
    Formats a box.
    :param left_margin:     Width of left margin 
    :param centre_margin:   Width of centre margin 
    :param right_margin:    Width of right margin 
    :param left_text:       Text in left column 
    :param right_text:      Text in right column 
    :return: 
    """
    r = []
    left_space = centre_margin - left_margin - 1
    right_space = right_margin - centre_margin
    
    left_lines = list( ansi_helper.wrap( left_text, left_space ) )
    right_lines = list( ansi_helper.wrap( right_text, right_space ) )
    num_lines = max( len( left_lines ), len( right_lines ) )
    
    for i in range( num_lines ):
        left = left_lines[i] if i < len( left_lines ) else ""
        right = right_lines[i] if i < len( right_lines ) else ""
        
        text = (" " * left_margin) + ansi_helper.ljust( left, left_space, " " ) + Style.RESET_ALL + " " + right + Style.RESET_ALL
        r.append( text )
    
    return "\n".join( r )
